#!C:\conda\envs\_build\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'conda-build-all==0.13.3','console_scripts','conda-buildall'
__requires__ = 'conda-build-all==0.13.3'
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.exit(
        load_entry_point('conda-build-all==0.13.3', 'console_scripts', 'conda-buildall')()
    )
